// WAVEFLT2 header

#ifndef _WFLT_FILTER_H
#define _WFLT_FILTER_H

#include "waveflt.h"
#include "wflt2dll.h"

#define VER_STR "WAVEFLT2 v.1.16 1/18/2003 (c)1999-2003 T.Tokairin\nMersenne Twister: (c) 1997-2002, M.Matsumoto and T.Nishimura"
#define VER_NUM  110

#define PI 3.14159265358979324

#define MAX_CHN		6   // max number of channel

#define QUANTIZATION(a,b) ((LONG)(a+b))

// ID of MDCT 
#define ID_MDCT_ADPMDCT 0


// ID of FFT
#define ID_FFT_ADPMDCT 0

//--------------------------------------------------
// adp.c
void PrepareADP(DWORD dwADPleng,  // length
				double dDB,   // loss
				double dbLevel, 
				DWORD dwBufferSize,
				DWORD dwChn
				);

void unprepareADP();
void ClearADPCoef();
void ADPLMS(double* lpFilterBuf, // buffer
			DWORD dwPointsInBuf, // points
			DWORD dwChn
			);




//----------------------------
// conv.c
double CONV();
void InitCONV(DWORD dwLength);
void UnprepareCONV();
void SetConvSize(DWORD n);
double* GetConvBufA();
double* GetConvBufB();



//--------------------------------------------------
// fir.c
void FIR(double* lpFilterBuf, // filter buffer
		 DWORD dwPointsInBuf, // points of data in the filter buffer
		 DWORD dwFirNum, // number of filter
		 DWORD dwChn // channel
		 );


void prepareFIRCoefficient(DWORD dwFilter,  // type of filter
						   DWORD dwFIRleng, // length of filter
						double dbSmpRate,  // sampling rate
						double dDB,  // dB, loss
						DWORD dwCutLow,  // cut-off freq (low)
						DWORD dwCutHigh,  // cut-off freq (high)
						DWORD dwBufSize, // buffer size
						DWORD dwFirNum, // number of filter
						DWORD dwChn // channel
						);

void CalcEQCoefficient(DWORD dwFIRleng,
					   DWORD dwBand,  // band
					   double dbSmpRate, // sampling rate
					   double dDB,  // dB, loss
					   double* dbEQlevel,  // dB, levels
					   double dQ, // Q
					   DWORD dwBufSize, // buffer size
					   DWORD dwFirNum, // number of filter
					   DWORD dwChn // channel
					   );

void ClearFirBuf();
void unprepareFIR();
void RewindBufFIR(DWORD dwRewSize,DWORD dwFirNum,DWORD dwChn);

void CalcFirCoefficient(DWORD dwFilter,  // type of filter
						double * coeff, 

						DWORD dwLength,
						double dbSmpRate,  // sampling rate
						double dDB,  // loss
						DWORD dwCutLow,  // cut-off freq (low)
						DWORD dwCutHigh  // cut-off freq (high)
						);




//--------------------------------------------------
// iir.c
void IIR(double* lpFilterBuf, // buffer
		 DWORD dwPointsInBuf, // points

		 DWORD dwIirNum,
		 DWORD dwChn
		 );


void prepareIIRCoefficient(DWORD dwFilter,// type of filter
						   double dCutFreqL, // low
						   double dCutFreqH, // hish
						   double dSmpRate,  // sampling rate
						   DWORD N,  // dimension

						   DWORD dwIirNum, // number of filter
						   DWORD dwChn // channel
						);

void CalcIirShelvingEQ(DWORD dwFilter,
					   double dCutFreq,
					 double dDb,
					 double dSmpRate,  // sampling rate
					 DWORD dwIirNum, // number of filter
					 DWORD dwChn // channel
					 );


void CalcIirPeakingEQ(double dCntFreq,  // center
					  double dQ,  // Q
					  double dDb,
					  double dSmpRate,  // sampling rate
					  DWORD dwIirNum, // number of filter
					  DWORD dwChn // channel
					  );

void ClearIirBuf();

void IIR2_Butterworth(DWORD dwFilter,// type of filter
					double a[3],
					double b[3],
					double k,
					double N,

					double dCutFreqL, // low
					double dCutFreqH, // hish
						
					double dSmpRate  // sampling rate
					);

void IIR4_BPF_Butterworth(double a[5],
					double b[5],
					double k,
					double N,

					double dCutFreqL, // low
					double dCutFreqH, // hish
						
					double dSmpRate  // sampling rate
					);

void IIR1_ShelvingEQ(DWORD dwFilter,
				   double a[2],
				   double b[2],
				   double dCutFreq,
				   double dDb,
				   double dSmpRate  // sampling rate
						   );


void IIR2_PeakingEQ(double a[3],
				   double b[3],

					double dCutFreqL, // low
					double dCutFreqH, // hish
					double dDb,

					double dSmpRate  // sampling rate
					);


//--------------------------------------------------
// wavesh.c
DWORD SHRINK_MDCT(double* lpFilterBuf, // filter buffer
			DWORD dwPointsInBuf, // points
			DWORD dwCh			// channel
			);

void unprepareShrink();
void ClearShrink();

void prepareShrink(DWORD dwShLeng,
				   WAVEFORMATEX waveFmt,
				   double dThreshold  // dB
				   );


//--------------------------------------------------
// rsmp.c
void prepareRsmp(DWORD dwInputFreq, DWORD dwOutputFreq, 
				 DWORD dwCutFreq, double dDB, 
				 DWORD dwLength, 
				 DWORD dwUp,DWORD dwDown,
				 DWORD dwFilterBufSize,DWORD dwChn);

void unprepareRsmp();
void ClearRsmp();
void RSAMP( double* lpFilterBuf, // filter buffer
		   DWORD dwPointsInBuf,
			DWORD dwCh,

			LPDWORD lpdwAfterPoints
			);


//--------------------------------------------------
// mixfile.c
void ClearMixFile();
void CloseMixFile();
void MixFile(WAVEFORMATEX waveOrgFmt, 
			 double* lpFilterBuf[2],
			 DWORD dwPoints, // points in buffer
			 LONGLONG n64OutSize, // output size of data
			 double dMixLevel[2],  // mixing level
			 double dMixStartTime // mixing start time
			 );

BOOL OpenMixFile(char* szMixFile,
				 WAVEFORMATEX waveOrgFmt, 
				 DWORD dwBufSize, // points of data in buffer
				 double dMixStartTime, // start time in mixing file
				 char* lpszErr
				 );

//--------------------------------------------------
// compr.c
void ClearCOMP();
void COMPRESS(long nSamplesPerSec,
			 double* lpFilterBuf[2],
			 DWORD dwPointsInBuf, 
			 DWORD dwChn,

			 double dTh,  // threshold
			 double dRatio,
			 double dAttackTime,  // msec
			 double dReleaseTime,  // msec
			 DWORD dwCompNum
			 );
void prepareCOMP(DWORD dwChn,DWORD dwRMSsize,DWORD dwCompNum);
void unprepareCOMP();



//--------------------------------------------------
// split.c
VOID SPLIT(WAVEFORMATEX waveFmt, 
		   DWORD* lpdwPointsInBuf, // points in buffer
		   LONGLONG n64TotalOutSize, // total output size
		   DWORD dwNoSndCurNo,
		   BOOL* bChangeFile, // if bChangeFile = true, then file is changed
		   LONGLONG* lpn64SplitByte, // if output size exceeds this size, then change file

		   double* lpdSplitTime, // for -split2,3
		   LONGLONG* lpn64SplitByteMalti // for -splitbm
		   );


VOID NOSOUND(double* lpFilterBuf[2], // buffer
			  DWORD* lpdwPointsInBuf, // points of data in buffer
			   BOOL* lpbChangeFile, // if *lpbChangeFile = true, change output file
			   HWND hWnd, // HWND of LockOn
			   long nSamplesPerSec,
			   DWORD dwChn,

			   DWORD dwNoSndTime,
			   DWORD dwNoSndM1toM2, 
			   DWORD dwNoSndRecStop, 

			   double dNoSndBoundHead, // level of sound (dB)
			   double dNoSndBound // level of no sound (dB)
			   );
void PrepareNOSOUND(DWORD dwNoSndMode, DWORD dwNoSndSuspend, long nSamplesPerSec);


//--------------------------------------------------
// dcoffset.c
VOID DCOFFSET(double* lpFilterBuf, // buffer
			  DWORD dwPointsInBuf, // points
			  double dOffset);

VOID AUTODCOFFSET(double* lpFilterBuf[2], // buffer (L-R)
			  DWORD dwPointsInBuf, // points
			  WAVEFORMATEX waveFmt,  // format
			  DWORD dwTrainSec // sec, training time
			  );

void ClearAutoDCOffset();

//--------------------------------------------------
// fft.c
void calcFftFast(COMPLEX* data,
			 DWORD dwFFTnum
			 );
void calcIfftFast(COMPLEX* data,
			 DWORD dwFFTnum
			 );
void prepareFft(DWORD dwFftPoint,
				DWORD dwFFTnum
				);
void unprepareFft();


//--------------------------------------------------
// mdct.c
void prepareMDCT(DWORD N,DWORD dwMDCTnum,DWORD dwFFTnum);
void unprepareMDCT();
void MDCT_fast(double* x, // N	
		  double* f,	// N/2	
		  DWORD dwMDCTnum,
		  DWORD dwFFTnum
		  );
void IMDCT_fast(double* x,	// N 
				double* f,	// N/2 
				DWORD dwMDCTnum,
				DWORD dwFFTnum
				);



//--------------------------------------------------
// fadeinout.c
void FADEINOUT(double* lpFilterBuf, // filter buffer
				   DWORD dwPointsInBuf, // points of data in buffer

				   WAVEFORMATEX waveFmt,
				   DWORD dwFadeIn, // points of fade in
				   DWORD dwFadeOut, // points of fase out

				   LONGLONG n64TotalOutSize, // current size of output
				   LONGLONG n64TotalSize // total size of input file
				   );


// normal.c
void ClearNormal();
void SET_PEAK(double* lpFilterBuf, // buffer
			   DWORD dwPointsInBuf, // points
			   DWORD dwChn // channel
			   );

void SET_AVG(double* lpFilterBuf, // buffer
			 DWORD dwPointsInBuf, // points
			 DWORD dwChn // channel
			 );

void SET_RMS(double* lpFilterBuf, // buffer
			 DWORD dwPointsInBuf, // points
			 DWORD dwChn // channel
			 );




//-------------------------------------------------
// filterfunc.c


double kaiserAlpha(double dDB);
double I0(double x);
void Idft(LPCOMPLEX data, // input
		  LPCOMPLEX out, // output
		  DWORD N  
		  );
void KBDWin(double* win, DWORD dwSize,double alpha);

void AddSinCurve(double* lpFilterBuf,
				 DWORD dwPointsInBuf,

				 DWORD dwRate,
				 DWORD dwChn,
				 DWORD dwNum, //number of waves  (if 0,  then initialize)
				 LPDWORD dwFreq,double* lpdbDb,LPDWORD dwPhase);

VOID StopRecording(HWND hWnd);


//------------------------------------------------
// fbank.c
void ClearFltBank();
VOID FltBank_analysis(double *dInput, // SUB_BAND point, input
				double *dOutput,   // SUB_BAND point, output
				DWORD dwCh);
// synthesis
VOID FltBank_synthesis(double *dOutput, // SUB_BAND point, output
				double *dInput,   // SUB_BAND point, input
				DWORD  dwCh // channel
				);


//--------------------------------------------
// ngate.c
void NOISEGATE_FREQ(double* lpFilterBuf[2], // filter buffer
			DWORD dwPointsInBuf, // points
			WAVEFORMATEX waveFmt
			);
void unprepareNGATE();
void ClearNGATE();
void prepareNGATE(WAVEFORMATEX waveFmt,
				  double dThreshold,  // dB
				  double dRelease, // msec
				  double dAttack, // msec
				  DWORD dwRMS
				   );

//--------------------------------------------
// dither.c
void ClearNSFBuf();
void DITHER(double* lpFilterBuf, // filter buffer
		 DWORD dwPointsInBuf, // points of data in the filter buffer
		 DWORD dwChn, // channel
		 double dAmp
		 );

//--------------------------------------------
// rand.c
void init_genrand(unsigned long s);
unsigned long genrand_int32(void);

#endif
