//-----------------------------------------------------------
// WAVEFLT2 
// Copyright (c) 1999-2002 Tomoya Tokairin
// LICENSE: GPL
//
// DWT , IDWT

#include "filter.h"

#define DWT_MAX_SUPP 20
#define MODE_DWT	  1
#define MODE_IDWT	  0
#define DWT_NUM		  4

typedef struct
{
	DWORD dwCofSize;
	DWORD dwOffset;
	double smCoef[DWT_MAX_SUPP];   // smooting
	double dtCoef[DWT_MAX_SUPP];   // detail
}QMFS,*LPQMFS;

QMFS DWT_Qmfs[DWT_NUM];	    	// QMFs
double *DWT_WorkMem[DWT_NUM]; 
DWORD 	DWT_dwLn[DWT_NUM];


//--------------------------------------------------------------------
// Partial Wavelet Transform
void PartialWaveletTrans(double *data, 
						 DWORD dwPoints,
						 int nMode,
						 LPQMFS lpQmfs,
						 double *dWorkMem
						 ){

	double dat1,dat2;
	DWORD i,i2,i3;
	DWORD dwBasePoint;
	DWORD dwCurPoint;  

	memset(dWorkMem,0,sizeof(double)*dwPoints);

	// DWT
	if(nMode == MODE_DWT)
	{   
		for(i = 0,i2 = 0; i < dwPoints; i+=2,i2++)
		{
			dwBasePoint = i + lpQmfs->dwOffset;

			// convolution
			for(i3 = 0; i3 < lpQmfs->dwCofSize; i3++)
			{
				dwCurPoint = (dwBasePoint + i3) & (dwPoints-1);

				// smooting
				dWorkMem[i2] += lpQmfs->smCoef[i3] * data[dwCurPoint];     

				// detail
				dWorkMem[dwPoints/2 + i2] += lpQmfs->dtCoef[i3] * data[dwCurPoint];  
			}

		}

	}
	// IDWT
	else
	{  
		for(i = 0,i2 = 0; i < dwPoints; i+=2,i2++)
		{
			dat1 = data[i2];  // smooting
			dat2 = data[dwPoints/2 + i2]; // detail
			dwBasePoint = i + lpQmfs->dwOffset;

			for(i3 = 0; i3 < lpQmfs->dwCofSize; i3++)
			{
				dwCurPoint = (dwBasePoint + i3) & (dwPoints-1);
				dWorkMem[dwCurPoint] += lpQmfs->smCoef[i3] * dat1;
				dWorkMem[dwCurPoint] += lpQmfs->dtCoef[i3] * dat2;
			}
		}
	}

	memcpy(data,dWorkMem,sizeof(double)*dwPoints);
}



//--------------------------------------------------------------------
// DWT 
// data:  low freq.  <- data[0], ... , data[DWT_dwLn-1] -> high freq.
DWORD DWT(double *data, 
		 DWORD dwDWTnum
		 )
{
	DWORD dwDataPoints,dwDataPoints2;

	dwDataPoints = dwDataPoints2 = DWT_dwLn[dwDWTnum];

	while(dwDataPoints2 > 1)
	{
		PartialWaveletTrans(data,dwDataPoints2,MODE_DWT
			,&DWT_Qmfs[dwDWTnum],DWT_WorkMem[dwDWTnum]);
		dwDataPoints2 /= 2;
	}

	return DWT_dwLn[dwDWTnum];
}



//--------------------------------------------------------------------
// IDWT 
DWORD IDWT(double *data,
		 DWORD dwDWTnum
		 )
{
	DWORD dwDataPoints,dwDataPoints2;

	dwDataPoints = DWT_dwLn[dwDWTnum];
	dwDataPoints2 = 2;

	while(dwDataPoints2 <= dwDataPoints)
	{
		PartialWaveletTrans(data,dwDataPoints2,MODE_IDWT
			,&DWT_Qmfs[dwDWTnum],DWT_WorkMem[dwDWTnum]);
		dwDataPoints2 *= 2;
	}

	return DWT_dwLn[dwDWTnum];
}



//--------------------------------------------------------------------
// set quadrature mirror filters 
BOOL setDaubechiesFilter(long nDaub, 
						 BOOL bCentering,  // centering
						 LPQMFS lpQmfs
						 )
{
	DWORD k;
	double sig = -1;

	lpQmfs->dwCofSize = nDaub;
	switch(nDaub)
	{
	case 4:	
		lpQmfs->smCoef[0] = 0.482962913145 ;
		lpQmfs->smCoef[1] = 0.836516303738 ;
		lpQmfs->smCoef[2] = 0.224143868041 ;
		lpQmfs->smCoef[3] = -0.129409522551 ;
		break ;
		
	case 6:	
		lpQmfs->smCoef[0] = 0.332670552950 ;
		lpQmfs->smCoef[1] = 0.806891509311 ;
		lpQmfs->smCoef[2] = 0.459877502118 ;
		lpQmfs->smCoef[3] = -0.135011020010 ;
		lpQmfs->smCoef[4] = -0.085441273882 ;
		lpQmfs->smCoef[5] = 0.035226291882 ;
		break ;
		
	case 8:	
		lpQmfs->smCoef[0] = 0.230377813309 ;
		lpQmfs->smCoef[1] = 0.714846570553 ;
		lpQmfs->smCoef[2] = 0.630880767930 ;
		lpQmfs->smCoef[3] = -0.027983769417 ;
		lpQmfs->smCoef[4] = -0.187034811719 ;
		lpQmfs->smCoef[5] = 0.030841381836 ;
		lpQmfs->smCoef[6] = 0.032883011667 ;
		lpQmfs->smCoef[7] = -0.010597401785 ;
		break ;
		
	case 10:
		lpQmfs->smCoef[0] = 0.160102397974 ;
		lpQmfs->smCoef[1] = 0.603829269797 ;
		lpQmfs->smCoef[2] = 0.724308528438 ;
		lpQmfs->smCoef[3] = 0.138428145901 ;
		lpQmfs->smCoef[4] = -0.242294887066 ;
		lpQmfs->smCoef[5] = -0.032244869585 ;
		lpQmfs->smCoef[6] = 0.077571493840 ;
		lpQmfs->smCoef[7] = -0.006241490213 ;
		lpQmfs->smCoef[8] = -0.012580751999 ;
		lpQmfs->smCoef[9] = 0.003335725285 ;
		break ;
		
	case 12:
		lpQmfs->smCoef[0] = 0.111540743350 ;
		lpQmfs->smCoef[1] = 0.494623890398 ;
		lpQmfs->smCoef[2] = 0.751133908021 ;
		lpQmfs->smCoef[3] = 0.315250351709 ;
		lpQmfs->smCoef[4] = -0.226264693965 ;
		lpQmfs->smCoef[5] = -0.129766867567 ;
		lpQmfs->smCoef[6] = 0.097501605587 ;
		lpQmfs->smCoef[7] = 0.027522865530 ;
		lpQmfs->smCoef[8] = -0.031582039318 ;
		lpQmfs->smCoef[9] = 0.000553842201 ;
		lpQmfs->smCoef[10] = 0.004777257511 ;
		lpQmfs->smCoef[11] = -0.001077301085 ;
		break ;
		
	case 14:
		lpQmfs->smCoef[0] = 0.077852054085 ;
		lpQmfs->smCoef[1] = 0.396539319482 ;
		lpQmfs->smCoef[2] = 0.729132090846 ;
		lpQmfs->smCoef[3] = 0.469782287405 ;
		lpQmfs->smCoef[4] = -0.143906003929 ;
		lpQmfs->smCoef[5] = -0.224036184994 ;
		lpQmfs->smCoef[6] = 0.071309219267 ;
		lpQmfs->smCoef[7] = 0.080612609151 ;
		lpQmfs->smCoef[8] = -0.038029936935 ;
		lpQmfs->smCoef[9] = -0.016574541631 ;
		lpQmfs->smCoef[10] = 0.012550998556 ;
		lpQmfs->smCoef[11] = 0.000429577973 ;
		lpQmfs->smCoef[12] = -0.001801640704 ;
		lpQmfs->smCoef[13] = 0.000353713800 ;
		break ;
		
	case 16:
		lpQmfs->smCoef[0] = 0.054415842243 ;
		lpQmfs->smCoef[1] = 0.312871590914 ;
		lpQmfs->smCoef[2] = 0.675630736297 ;
		lpQmfs->smCoef[3] = 0.585354683654 ;
		lpQmfs->smCoef[4] = -0.015829105256 ;
		lpQmfs->smCoef[5] = -0.284015542962 ;
		lpQmfs->smCoef[6] = 0.000472484574 ;
		lpQmfs->smCoef[7] = 0.128747426620 ;
		lpQmfs->smCoef[8] = -0.017369301002 ;
		lpQmfs->smCoef[9] = -0.044088253931 ;
		lpQmfs->smCoef[10] = 0.013981027917 ;
		lpQmfs->smCoef[11] = 0.008746094047 ;
		lpQmfs->smCoef[12] = -0.004870352993 ;
		lpQmfs->smCoef[13] = -0.000391740373 ;
		lpQmfs->smCoef[14] = 0.000675449406 ;
		lpQmfs->smCoef[15] = -0.000117476784 ;
		break ;
		
	case 18:
		lpQmfs->smCoef[0] = 0.038077947364 ;
		lpQmfs->smCoef[1] = 0.243834674613 ;
		lpQmfs->smCoef[2] = 0.604823123690 ;
		lpQmfs->smCoef[3] = 0.657288078051 ;
		lpQmfs->smCoef[4] = 0.133197385825 ;
		lpQmfs->smCoef[5] = -0.293273783279 ;
		lpQmfs->smCoef[6] = -0.096840783223 ;
		lpQmfs->smCoef[7] = 0.148540749338 ;
		lpQmfs->smCoef[8] = 0.030725681479 ;
		lpQmfs->smCoef[9] = -0.067632829061 ;
		lpQmfs->smCoef[10] = 0.000250947115 ;
		lpQmfs->smCoef[11] = 0.022361662124 ;
		lpQmfs->smCoef[12] = -0.004723204758 ;
		lpQmfs->smCoef[13] = 0.004281503682 ;
		lpQmfs->smCoef[14] = 0.001847646883 ;
		lpQmfs->smCoef[15] = 0.000230385764 ;
		lpQmfs->smCoef[16] = -0.000251963189 ;
		lpQmfs->smCoef[17] = 0.000039347320 ;
		break ;
		
	default:
	case 20:
		lpQmfs->smCoef[0] = 0.026670057901 ;
		lpQmfs->smCoef[1] = 0.188176800078 ;
		lpQmfs->smCoef[2] = 0.527201188932 ;
		lpQmfs->smCoef[3] = 0.688459039454 ;
		lpQmfs->smCoef[4] = 0.281172343661 ;
		lpQmfs->smCoef[5] = -0.249846424327 ;
		lpQmfs->smCoef[6] = -0.195946274377 ;
		lpQmfs->smCoef[7] = 0.127369340336 ;
		lpQmfs->smCoef[8] = 0.093057364604 ;
		lpQmfs->smCoef[9] = -0.071394147166 ;
		lpQmfs->smCoef[10] = -0.029457536822 ;
		lpQmfs->smCoef[11] = 0.033212674059 ;
		lpQmfs->smCoef[12] = 0.003606553567 ;
		lpQmfs->smCoef[13] = -0.010733175483 ;
		lpQmfs->smCoef[14] = 0.001395351747 ;
		lpQmfs->smCoef[15] = 0.001992405295 ;
		lpQmfs->smCoef[16] = -0.000685856695 ;
		lpQmfs->smCoef[17] = -0.000116466855 ;
		lpQmfs->smCoef[18] = 0.000093588670 ;
		lpQmfs->smCoef[19] = -0.000013264203 ;
		lpQmfs->dwCofSize = 20;
		break ;
	}

	// detail
	for(k = 0; k < lpQmfs->dwCofSize; k++)
	{
		lpQmfs->dtCoef[(lpQmfs->dwCofSize-1) - k] = sig * lpQmfs->smCoef[k];
		sig *= -1;
	}

	// centering
	if(bCentering) lpQmfs->dwOffset = lpQmfs->dwCofSize/2;
	else lpQmfs->dwOffset = 0;

	return TRUE;
}



//------------------------------------------------------
// unprepare
void unprepareDWT()
{
	DWORD i;
	for(i=0;i<=DWT_NUM;i++) if(DWT_WorkMem[i]) free(DWT_WorkMem[i]);
}


//------------------------------------------------------
// prepare
void prepareDWT(DWORD dwPoints,
				long nDaub, 
				DWORD dwDWTnum
				)
{
	if(DWT_WorkMem[dwDWTnum] == NULL)
	{
		DWT_WorkMem[dwDWTnum] = (double*)malloc(sizeof(double)*dwPoints +1024);
		setDaubechiesFilter(nDaub,false,&DWT_Qmfs[dwDWTnum]);
		DWT_dwLn[dwDWTnum] = dwPoints;
	}
}


// EOF
