//--------------------------------
// header of wflt2.dll
//
// Copyright (c) 1999-2002 Tomoya Tokairin
// LICENSE:  LGPL2.1

#ifndef _WFLT_WFLT2DLL_H
#define _WFLT_WFLT2DLL_H


#ifdef WFLT2DLL_EXPORTS
#define	EXPORT __declspec(dllexport) WINAPI
#else
#define	EXPORT WINAPI
#endif


#ifndef WAVE_FORMAT_PCM
#define WAVE_FORMAT_PCM 0x0001
#endif
#ifndef WAVE_FORMAT_IEEE_FLOAT
#define WAVE_FORMAT_IEEE_FLOAT 0x0003
#endif


#define MAX_SPLIT	128 // max of split number
#define MAX_SINE_WAVES	16 // max of sine waves
#define MAXMDCTFILTERPOINT 8192  // max of  MDCT filter length
#define MAXFILTERPOINT	8191  // max of  FIR filter length
#define EQBANDS	10 // max band of EQ 
#define MAX_EQDB	24 // dB, max level of EQ

#define WAVEHDRSIZE(a) ( (a == true) ? 44 + (8 + 8 + 4) : 44)

// type of filter
#define NO_FILTER	0
#define LPF 1
#define HPF 2
#define BPF 3
#define BSF 4
#define SVEQL 5  // shelving EQ low
#define SVEQH 6  // shelving EQ high


// mode of nosound
#define NOSND_NOT			0	// no split
#define NOSND_EXEC			1	// searching no sound part
#define NOSND_SEARCH		2	// searching sound part after splitting, then move to NOSND_EXEC
#define NOSND_CUT			3	// deleting no sound part after splitting, then move to NOSND_EXEC
#define NOSND_HEADCUTONLY	4	// deleting no sound part,then move to NOSND_NOT 
#define NOSND_STOP			5	// stop recording of lockon if no sound part is continuing

// type of normalizer
#define NORMAL_NOT			0
#define NORMAL_EXEC			1  // execute normalizer
#define NORMAL_PEAKBYDB		2  // searching peak
#define NORMAL_AVG			3  // searching average
#define NORMAL_RMS			4  // searching RMS


// ID of FIR filter
#define ID_FIR_NORMAL	0  // FIR
#define ID_FIR_EQ		1  // FIR-EQ
#define ID_FIR_NOSND	2  // FIR before nosound


// ID of IIR filter
#define ID_IIR_NORMAL	0  // IIR
#define ID_IIR_DEMP		1  // IIR-(de)emphasys
#define ID_IIR_SVEQL	2  // IIR-shelving EQ low
#define ID_IIR_SVEQH	3  // IIR-shelving EQ high
#define ID_IIR_PKEQ		4  // IIR-peaking EQ

#ifndef COMPLEX
typedef struct
{
	double r;
	double i;
}COMPLEX,*LPCOMPLEX;
#endif


//----------------------------------------
// parameters
// note: you need to call SetDefaultOption() to set default value before filtering.

typedef struct
{
	
	// DC offset adjustment
	BOOL bOffset; 
	double dOffset[2]; // offset value (left,right)

	// Auto DC offset adjustment
	BOOL bAutoOffset; 
	DWORD dwAutoOffsetTime; // sec, training time

	// split at no sound part
	DWORD dwNoSndMode; // initial mode. if NOSND_NOT, don't split
	DWORD  dwNoSndTime; // msec, time of no sound part
	double dNoSndBoundHead; // db, level of sound. when level of input value execeed this value, data goes into the sound part.
	double dNoSndBound; // db, level of no sound, when level of input value is under this value, data goes into the no sound part.
	DWORD dwNoSndSuspend; // msec. when split mode enter NOSND_EXEC, don't split file in dwNoSndSuspend msec.
	DWORD dwNoSndM1toM2; // msec. when time of no sound execeed this, change mode from NOSND_SEARCH to NOSND_CUT.
	DWORD dwNoSndRecStop; // msec. when time of no sound execeed this, stop recording of 'lockon'

	// FIR filter befor split
	DWORD dwNoSndFIRFilter; // type of filter, if NO_FILTER, don't use FIR
	DWORD dwNoSndFIRCutLow; // cut-off (low)
	DWORD dwNoSndFIRCutHigh; // cut-off (high)
	DWORD dwNoSndFIRleng; // length
	double dNoSndFIRDb; // db, loss of filter

	// split at specified size
	BOOL bSplit; 
	LONGLONG n64SplitByte; // byte,  if output size execeed this value, then split the file.
	double dSplitTime[MAX_SPLIT]; // sec, times
	LONGLONG n64SplitByteMalti[MAX_SPLIT]; // byte, size specified -splitbm 

	
	// LR-mixing
	BOOL bMixLR; 
	double dMixLRLevel; // level 0 << 1

	
	// file mixing
	BOOL bMixFile;
	char szMixFile[MAX_PATH];
	double dMixLevel[2]; // level 0< < 1
	double dMixStartTime[2]; // sec

	
	// synthesize sine waves
	DWORD dwAddSinNum;  // if = 0, don't synthesize
	DWORD dwAddSinFreq[MAX_SINE_WAVES];
	DWORD dwAddSinPhase[MAX_SINE_WAVES];
	double dAddSinDb[MAX_SINE_WAVES];

	
	// ADP
	BOOL bADP;
	DWORD dwADPDelayTime; // msec, delay time 
	double dADPlevel; // level
	DWORD dwADPleng; // length
	double dADPDb;  // db, loss of adp filter

	
	// FIR 
	DWORD dwFIRFilter; // type of filter, if NO_FILTER, don't use FIR
	DWORD dwFIRCutLow; // cut-off (low)
	DWORD dwFIRCutHigh; // cut-off (high)
	DWORD dwFIRleng; // length
	double dFIRDb; // db, loss of filter

	
	// FIR-EQ
	BOOL bFIREQ;
	DWORD dwEQband; // bands
	double dEQLevel[EQBANDS]; // dB, level
	DWORD dwEQleng; // length
	double dEQDb; // db, loss of filter
	double dEQ_Q;  // Q  

	
	// IIR 
	DWORD dwIIRFilter;  // type of filter. if NO_FILTER, don't filtering
	DWORD dwIIRCutLow; // cut-off freq (low)
	DWORD dwIIRCutHigh; // cut-off freq (high)

	// (de)emphasys
	BOOL bDemp;
	double dDempDb;  // db, 10 or -10 , t1 = 0.000050, t2 = 0.000015 are const.

	// shelving EQ low
	BOOL bSVEQL;
	double dSVEQLdb;
	DWORD dwSVEQLfreq;

	// shelving EQ high
	BOOL bSVEQH;
	double dSVEQHdb;
	DWORD dwSVEQHfreq;

	// peaking EQ
	BOOL bPKEQ;
	double dPKdb;
	double dPKQ;
	DWORD dwPKfreq;

	// waveshrink filter
	BOOL bWaveSh;
	double dWaveShLevel;
	double dWaveShLowLevel;
	double dWaveShHiLevel;
	DWORD dwWaveShleng; // length

	// noise gate
	BOOL bNgate;
	double dNgateTh;   // db, threshold
	double dNgateRelease; // msec
	double dNgateAttack; // msec
	DWORD dwNgateRMS;  // RMS
	
	// phase invert
	BOOL bPhaseInv;
	
	// compressor 
	BOOL bComp;
	double dCompRatio; // ratio
	double dCompTh; // threshold (db)
	double dCompAttack; // attack time (msec)
	double dCompRelease; // release time (msec)
	DWORD dwCompRMS;  // RMS flame length 
	
	
	// limiter atter normaizing
	BOOL bNormalUseCompressor;  
	double dNormalRatio; // ratio
	double dNormalTh; // threshold (db)
	double dNormalAttack; // attack time (msec)
	double dNormalRelease; // release time (msec)
	DWORD dwNormalCompRMS; // RMS flame length 
	
	// volume
	double dVolume; // >= 0, if =1.0 , don't adjust gain
	
	// balance
	BOOL bBalance;
	double dBalance[2]; // L-R, >= 0

	
	// fade in/out
	DWORD dwFadeIn; // byte, size of fade in
	DWORD dwFadeOut; // byte, size of fade out
	
	
	// re-sampling
	BOOL bRsmp;
	DWORD dwRsmpInFreq; // input freq
	DWORD dwRsmpOutFreq; // output freq
	DWORD dwRsmpUp;  // up-sampling  points
	DWORD dwRsmpDown; // down-sampling points after up-sampling
	DWORD dwRsmpLn; // length of FIR-LPF filter
	DWORD dwRsmpCut; // cut-off freq of FIR-LPF filter
	double dRsmpDB; // db, loss of FIR-LPF filter

	
	// change bits of output
	DWORD dwBitChange; // = 0,8,16,32,64. if 0, don't change
	DWORD dwBitChangeTag; // tag
	
	// dither
	BOOL bDither;
	double dDitherAmp; // amp
	
}FILTER_DATA,*LPFILTER_DATA;


//--------------------------------------------------------------------------
// functions 
//--------------------------------------------------------------------------

//--------------------------------------------------------------------------
// basic functions


DWORD EXPORT GetVer(char* szVer); // get version
BOOL EXPORT CheckSSE2(); // you need to call this function to use SSE2


// filter.c
void EXPORT CopyBufferBtoD(BYTE* lpBuffer,  // input, buffer (BYTE*)
						DWORD dwByte, // size of lpBuffer

						double* lpFilterBuf[2], // output, buffer of wave data (double*) (L-R)
						LPDWORD lpdwPointsInBuf, // points of data in lpFilterBuf
						WAVEFORMATEX waveFmt
						);

void EXPORT CopyBufferDtoB(BYTE* lpBuffer,  // output, buffer(BYTE*)
						   
						   double* lpFilterBuf[2], // input, buffer(double*)
						   DWORD dwPointsInBuf, // points in lpFilterBuf
						   WAVEFORMATEX waveFmt,
   						   BOOL bRound,
						   
						   DWORD* lpdwSaturate  // number of saturation
						   );

void EXPORT ClearAllFilters();
void EXPORT UnprepareAllFilters();

BOOL EXPORT InitFilters(LPFILTER_DATA lpFDat,
				 DWORD dwFilterPoints, // points of data in filter buffer
				 WAVEFORMATEX inWaveFmt,
				 WAVEFORMATEX outWaveFmt,
				 char* lpszErr);


void EXPORT SetDefaultOption(LPFILTER_DATA lpFDat);

void EXPORT WFLT_FILTER(LPFILTER_DATA lpFDat,  // parameter

				 double* lpFilterBuf[2],  // (input / output) buffer of wave data
				 LPDWORD lpdwPointsInBuf, // (input / output) data points in buffer
				 LPDWORD lpdwRealPointsInBuf, // (output) output data points
				 // note: if inWaveFmt.nSamplesPerSec = outWaveFmt.nSamplesPerSec,
				 // then *lpdwPointsInBuf = *lpdwRealPointsInBuf.

				 BOOL* lpbChangeFile, // if return value of *lpbChangeFile is true, change output file
				 BOOL bADPtrainMode, // if true , now ADP is in the training mode  
				 DWORD dwCurrentNormalMode, // current mode of normalizer
				 double dNormalGain[2], // L-R, gain for normalizer
				 DWORD dwCurFileNo,  // current file number
				 HWND hWndLockon, // hwnd of 'lockon'

				 LONGLONG n64OutSize, // byte, total output size
				 LONGLONG n64DataSize, // byte, total data size of wave
				 WAVEFORMATEX inWaveFmt, // format of input
				 WAVEFORMATEX outWaveFmt // format of output
				 );


// wave.c
BOOL EXPORT CheckWaveFormat(LPWAVEFORMATEX lpWaveFmt,char* lpszErr);
// check whether format is valid or not


BOOL EXPORT GetWaveFormat(char* lpszFileName, // file name or 'stdin'
				   LPWAVEFORMATEX lpWaveFmt, 
				   LONGLONG* lpn64WaveDataSize, // size of data
				   LONGLONG* lpn64WaveOffset, // offset to data chunk
				   char* lpszErr 
				   );

VOID EXPORT WriteWaveHeader(HANDLE hdWriteFile,LPWAVEFORMATEX lpWaveFmt,LONGLONG n64WaveDataSize
					 ,BOOL bUseExtChunk);

void EXPORT SetWaveFmt(LPWAVEFORMATEX lpWaveFmt,WORD waveChn,
				DWORD waveRate,	WORD waveBit, WORD wTag);

LONGLONG EXPORT SeekStdin(LPBYTE lpBuffer,
			   DWORD dwBufSize,
			   LONGLONG n64SeekPointer,
			   LONGLONG n64CurFilePointer  // dwSeekPointer > dwCurFilePointer
			   );

void EXPORT WaveLevel(double dLevel[2],  // output of left and right
			   BYTE* lpWaveData,  // input
			   WAVEFORMATEX waveFmt);
// change byte-data to double-data 

double EXPORT GetMaxWaveLevel(WAVEFORMATEX waveFmt);
// get max level of wave 



//-----------------------
// other functions


// adp.c
void EXPORT ClearADPBuf();
DWORD EXPORT GetADPCoef(double* coef[2]);

// dcoffset.c
void EXPORT GetAutoOffset(double dOffset[2]);

// filterfunc.c
double EXPORT CalcSpec(LPCOMPLEX x,// input
				DWORD N, // number of data
				double freq,  // frequency
				double samp_freq // sampling frequency
				); 


// fir.c
DWORD EXPORT GetFIRCoef(double** coef,DWORD dwFirNum,DWORD dwChn);


// iir.c
DWORD EXPORT GetIIRimpulse(double* imp,DWORD N,DWORD dwIirNum,DWORD dwChn);
DWORD EXPORT GetIIRCoef(double* dCoef[2],DWORD dwIirNum,DWORD dwChn);


// iireq.c
DWORD EXPORT GetIIREQimpulse(double* imp,DWORD N,DWORD dwChn);


// normal.c
void EXPORT GET_RMS(double* dRMS);
void EXPORT GET_PEAK(double* dPeak);
void EXPORT GET_AVG(double* dAVG);


// rsmp.c
DWORD EXPORT GetRSMPCoef(double** h,DWORD dwInputFreq,LPDWORD lpdwUp);


// split.c
VOID EXPORT ClearNOSOUND();  
// you need to call this function after spliting file


#endif
