// This is the example of using wflt2.dll. (VC++6.0) 
//
// For more details, see wflt2dll.h (and source code of waveflt2.exe).
//
// LICENSE: this example itself is PDS, but 'wflt2.dll' is LGPL2.1

#include <windows.h>
#include <stdio.h>
#include "wflt2dll.h"

// usage: sample.exe lpf-cutoff input.wav output.wav
int main(int argc,char *argv[]){

	char szVer[256],szInput[MAX_PATH],szOutput[MAX_PATH];
	HANDLE hdReadFile,hdWriteFile;
	LARGE_INTEGER LI;

	WAVEFORMATEX waveFmt,outwaveFmt;
	LONGLONG n64Offset,n64DataSize,n64OutSize;

	BYTE* lpBuffer;
	double* lpFilterBuffer[2];
	DWORD dwBufSize,dwOutBufSize; // bytes, buffer size of 'lpBuffer'
	DWORD dwFilterSize; // max points of 'lpFilterBuffer'
	// dwFilterSize = dwBufSize /= waveFmt.nBlockAlign
	
	FILTER_DATA fdata; // FILTER_DATA is defined in wflt2dll.h
	DWORD dwPointsInBuf,dwOutPointsInBuf; // points of data in 'lpFilterBuffer'

	DWORD i,dwByte;

	strcpy(szInput,argv[2]);
	strcpy(szOutput,argv[3]);
	dwBufSize = 256 * 1024;

	// get version
	GetVer(szVer);
	fprintf(stderr,"%s\n",szVer);	

	fprintf(stderr,"LPF cut-off freq. = %s hz\n",argv[1]);

	// use SSE2, if possible
	CheckSSE2();

	// get wave format of input file
	GetWaveFormat(szInput,&waveFmt,&n64DataSize,&n64Offset,NULL);
	outwaveFmt = waveFmt;

	fprintf(stderr,"%d hz, %d bit, %d ch\n",
		waveFmt.nSamplesPerSec, waveFmt.wBitsPerSample, waveFmt.nChannels);
	fprintf(stderr,"size = %d byte, offset = %d byte\n",(DWORD)n64DataSize,(DWORD)n64Offset);

	// make buffer
	lpBuffer = (BYTE*)malloc(sizeof(BYTE)*dwBufSize);

	dwFilterSize = dwBufSize/waveFmt.nBlockAlign;
	for(i=0;i<waveFmt.nChannels;i++) 
		lpFilterBuffer[i] = (double*)malloc(sizeof(double)*dwFilterSize);

	// set default values of parameters
	// note: you have to call this function before InitFilters()
	SetDefaultOption(&fdata);

	// use FIR Low Pass Filter
	fdata.dwFIRFilter = LPF;
	fdata.dwFIRCutLow = atoi(argv[1]);  // hz

	// if you want to use another filters, set parameters here before InitFilters()

	// initialize filter
	InitFilters(&fdata,dwFilterSize,waveFmt,outwaveFmt,NULL);

	// clear filter,  you have to call this function before filtering
	ClearAllFilters();

	// open files
	hdReadFile = CreateFile(szInput,GENERIC_READ, 0, 0,OPEN_EXISTING, FILE_ATTRIBUTE_READONLY, NULL);
	hdWriteFile = CreateFile(szOutput,GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL); 

	// move file pointers
	LI.QuadPart = n64Offset;
	SetFilePointer(hdReadFile,LI.LowPart, &LI.HighPart,FILE_BEGIN);
	SetFilePointer(hdWriteFile,44,NULL,FILE_BEGIN);

	n64OutSize = 0;  
	for(;;){

		// read
		ReadFile(hdReadFile,lpBuffer,dwBufSize,&dwByte,NULL);
		if(dwByte == 0) break; // EOF

		// byte -> double
		CopyBufferBtoD(lpBuffer,dwByte,lpFilterBuffer,&dwPointsInBuf,waveFmt);

		// note: dwPointsInBuf = dwByte / waveFmt.nBlockAlign

		// filter
		WFLT_FILTER(&fdata,
			lpFilterBuffer,
			&dwPointsInBuf, 
			&dwOutPointsInBuf, 
			NULL,FALSE,NORMAL_NOT,NULL,0,NULL,
			n64OutSize,n64DataSize,waveFmt,outwaveFmt);

		// if outwaveFmt.nSamplesPerSec = waveFmt.nSamplesPerSec, (i.e. resampling is 
		// not executed), then dwPointsInBuf = dwOutPointsInBuf.

		// double -> byte
		CopyBufferDtoB(lpBuffer,lpFilterBuffer,dwOutPointsInBuf,outwaveFmt,NULL);
		dwOutBufSize = dwOutPointsInBuf * outwaveFmt.nBlockAlign;

		// write
		WriteFile(hdWriteFile,lpBuffer,dwOutBufSize,&dwByte, NULL);
		n64OutSize += dwByte;

		fprintf(stderr,"\015%d k   ",(DWORD)n64OutSize/1024);

	}

	// write wave header (44 byte)
	WriteWaveHeader(hdWriteFile,&outwaveFmt,n64OutSize,FALSE);

	// close all filters
	UnprepareAllFilters();

	// close
	CloseHandle(hdReadFile);
	CloseHandle(hdWriteFile);
	free(lpBuffer);
	for(i=0;i<waveFmt.nChannels;i++) free(lpFilterBuffer[i]);

	return 0;
}








